// Copyright (c) 2005  John Abbott
// This file is part of the CoCoALib suite of examples.
// You are free to use any part of this example in your own programs.

#include "CoCoA/library.H"
using namespace CoCoA;
using namespace std;

//----------------------------------------------------------------------
const string ShortDescription =
  "Example program illustrating the creation of matrices, and some \n"
  "some operations on them. \n";

const string LongDescription =
  "Please note that the matrix code is still rather young. \n";
//----------------------------------------------------------------------

// Includes from the standard C++ library
// #include <iostream> // using std::endl;


// convention: a function containing a "new" should be named "New.."
matrix NewMatrixFromC(ring Z, int cmat[4][4])
{
  matrix M(NewDenseMat(Z,4,4));
  
  for (int i=0; i < 4; ++i)
    for (int j=0; j < 4; ++j)
      SetEntry(M, i, j, cmat[i][j]);
  return M;
}


void program()
{
  GlobalManager CoCoAFoundations;

  GlobalOutput() << ShortDescription << endl;

  const ring Z101 = NewZmod(101);
  const ring Q = RingQ();

   // the first 2 rows represent the degree matrix
  int M[4][4] = {{2, 0, 0, 3},
                 {0, 0, 8, 0},
                 {1, 2, 4, 1},
                 {0, 1, 0, 0}};

  // Create two CoCoALib matrices containing the values of M.
  // M_Z101 contains the image of M in Z/(101), and M_Q the image in Q.
  matrix M_Z101(NewMatrixFromC(Z101, M));
  matrix M_Q(NewMatrixFromC(Q, M));

  GlobalOutput() << "M_Z101 = " << M_Z101 << endl;
  matrix InvM_Z101 = inverse(M_Z101);  
  GlobalOutput() << "InvM_Z101 = " << InvM_Z101 << endl;
  mul(InvM_Z101, InvM_Z101, M_Z101);  
  GlobalOutput() << "InvM_Z101 * M_Z101 = " << InvM_Z101 << endl;

  GlobalOutput() << "----------------------------------" << endl;

  GlobalOutput() << "M_Q = " << M_Q << endl;
  GlobalOutput() << "rank(M_Q) = " << rank(M_Q) << endl;
  GlobalOutput() << "det(M_Q) = " << det(M_Q) << endl;
  matrix InvM_Q = inverse(M_Q);  
  GlobalOutput() << "InvM_Q = " << InvM_Q << endl;  
  matrix AdjM_Q = adjoint(M_Q);  
  GlobalOutput() << "AdjM_Q = " << AdjM_Q << endl;  
  mul(AdjM_Q, AdjM_Q, M_Q);  
  GlobalOutput() << "AdjM_Q * M_Q = " << AdjM_Q << endl;  

  GlobalOutput() << "----------------------------------" << endl;

  // It is easy to compute the determinant of an identity matrix, even if large.
  ConstMatrixView Id = IdentityMat(RingZ(), 4000);
  GlobalOutput() << "det(Id) = " << det(Id) << endl;  
}


// Use main() to handle any uncaught exceptions and warn the user about them.
int main()
{
  try
  {
    program();
    return 0;
  }
  catch (const CoCoA::ErrorInfo& err)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT CoCoA error";
    ANNOUNCE(err);
  }
  catch (const std::exception& exc)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT std::exception: " << exc.what() << endl;
  }
  catch(...)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT UNKNOWN EXCEPTION" << endl;
  }
  return 1;
}

//----------------------------------------------------------------------
// RCS header/log in the next few lines
// $Header: /Volumes/Home/cocoa/cvs-repository/CoCoALib-0.99/examples/ex-matrix1.C,v 1.3 2008/04/18 15:35:57 abbott Exp $
// $Log: ex-matrix1.C,v $
// Revision 1.3  2008/04/18 15:35:57  abbott
// (long overdue) Major revision to matrices
//
// Revision 1.2  2008/04/08 15:26:42  abbott
// Major revision to matrix implementation: added matrix views.
// Lots of changes.
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.6  2007/03/03 14:15:45  bigatti
// -- "foundations" renamed into "GlobalManager"
//
// Revision 1.5  2007/02/26 17:40:34  bigatti
// -- getting ready for unique ring Z: using NewZmod(N), NewRingQ()
//
// Revision 1.4  2007/02/22 17:25:40  bigatti
// -- added printing of separators: "-------"
//
// Revision 1.3  2007/02/12 16:27:43  bigatti
// -- added strings ShortDescription and LongDescription for indexing
//
// Revision 1.2  2007/02/10 18:44:03  cocoa
// Added "const" twice to each test and example.
// Eliminated dependency on io.H in several files.
// Improved BuildInfo, and added an example about how to use it.
// Some other minor cleaning.
//
// Revision 1.1.1.1  2006/05/30 11:39:36  cocoa
// Imported files
//
// Revision 1.1.1.1  2005/10/17 10:46:53  cocoa
// Imported files
//
// Revision 1.2  2005/10/14 15:25:07  cocoa
// Major tidying and cleaning to small prime finite fields.
// Several consequential changes.  Improved their documentation.
//
// Added Makefile and script to include/CoCoA/ directory to
// keep library.H up to date.
//
// Revision 1.1.1.1  2005/05/03 15:47:30  cocoa
// Imported files
//
// Revision 1.6  2005/04/27 16:14:56  cocoa
// Cleaned up example programs -- added "free use" permit.
// Changed a couple of ErrorInfo object names, and added
// ERR::NotTrueGCDDomain.
//
// Revision 1.5  2005/04/21 15:12:19  cocoa
// Revised NewPolyRing as Dag Arneson suggested (perhaps just an interim
// measure).
// Brought example programs up to date (new name for CoCoA error
// information objects).
//
// Revision 1.4  2005/04/19 14:06:04  cocoa
// Added GPL and GFDL licence stuff.
//
// Revision 1.3  2005/03/30 15:30:43  cocoa
// -- added: computation of rank and determinant
// -- changed: declaration of I4000 as ConstMatrix
//
// Revision 1.2  2005/03/02 18:46:41  cocoa
// Added new types ConstRefMatrix, and RefMatrix following along
// the lines of ConstRefRingElem and RefRingElem.  The semantics
// should be a bit clearer now.
//
// Revision 1.1.1.1  2005/01/27 15:12:13  cocoa
// Imported files
//
// Revision 1.1  2004/11/29 16:50:28  cocoa
// -- first import
//
