//   Copyright (c)  1997-2006  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License (version 2)
//   as published by the Free Software Foundation.  A copy of the full
//   licence may be found in the file COPYING in this directory.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoA; if not, write to the Free Software
//   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <stdio.h>
#include "jtime.h"
#include <math.h>

#include "Zdet.h"
#include "mpz_log.h"
#include "logi.h"
#include "jalloc.h"
#include "FF.h"
#include "FFdet.h"
#include "mpz_cra_ui.h"
#include "primes.h"
#include "Zsolve.h"
#include "Zmat_det_bound.h"


/***************************************************************************/

void Zdet(mpz_t det, Zmat M)
{
  Zmat rhs, soln;
  mpz_t *denom, D, det_factor;
  int i, j, r, singular;
  int row, col;
  int n = M->nrows;

printf("Full matrix Hadamard bound: %8.1f bits\n", Zmat_det_bound(M)/log(2.0));
  /* "Eliminate" random row and column by swapping with last row/col */
  /* We will swap them back at the end. */
  row = random()%n;
  col = random()%n;
  Zmat_swap_rows(M, row, n-1);
  Zmat_swap_cols(M, col, n-1);

  /* We are going to solve an (n-1)x(n-1) system with r=1 rhs vector */
  r = 1;
  rhs = Zmat_ctor(n-1, r);
  soln = Zmat_ctor(n-1, r);
  denom = (mpz_t*)MALLOC(r*sizeof(mpz_t));
  for(j=0; j < r; j++) mpz_init(denom[j]);
  for(i=0; i < n-1; i++) mpz_set(rhs->entry[i][0], M->entry[i][n-1]);

M->nrows=n-1;M->ncols=n-1;
  singular = Zsolve(soln, denom, M, rhs);
M->nrows=n;M->ncols=n;
  if (singular == 0) { printf("System was singular.\n"); exit(99); } /* BUG */
  /* MISSING CODE TO HANDLE SINGULAR CASE */

  mpz_init_set(D, denom[0]);
  mpz_init(det_factor);
  mpz_mul(det_factor, M->entry[n-1][n-1], denom[0]);
  { /* use a block to limit the scope of prod */
    mpz_t prod;
    mpz_init(prod);
    for (i=0; i < n-1; i++)
    {
      mpz_mul(prod, soln->entry[i][0], M->entry[n-1][i]);
      mpz_sub(det_factor, det_factor, prod);
    }
    mpz_clear(prod);
  }
  /* free rhs, soln, and denom */
  for(j=0; j < r; j++) mpz_clear(denom[j]);
  FREE(denom);
  Zmat_dtor(rhs);
  Zmat_dtor(soln);
printf("Solved linear system: ");print_time();
printf("D=");mpz_out_str(stdout,10,D);printf("\n");
  mpz_set_ui(det, 0);
  if (mpz_sgn(det_factor) == 0) goto tidy_up;

printf("Factor from linear system:  %8.1f bits\n", mpz_log(det_factor)/log(2.0));

  {
    PrimeSource PS = PrimeSourceCtor();
    FFmat Mp = FFmat_ctor(n-1, n-1);
    FFelem p, detp, Dp;
    int det_changed;
    FF FFp;
    mpz_t det_modulus;
    double log_det_modulus, det_bound;

    det_bound = 1 + Zmat_det_bound2(M, n-1); /* add 1 for safety */
printf("Submatrix Hadamard bound:   %8.1f bits\n", det_bound/log(2.0));
printf("Estimate of det of submat:  %8.1f bits\n", mpz_log(D)/log(2.0));
  det_bound -= mpz_log(D);
printf("Remaining distance to lift: %8.1f bits\n", det_bound/log(2.0));

    mpz_init_set_ui(det_modulus, 1);
    log_det_modulus = 0;

  while (log_det_modulus < det_bound)
  {
    p = NextPrime(PS);
    Dp = mpz_fdiv_ui(D, p);
    if (Dp == 0) continue; /* prime is "bad" so skip it */
    FFp = FFctor(p);
    FFselect(FFp);
    for (i=0; i < n-1; i++)
      for (j=0; j < n-1; j++)
        Mp->entry[i][j] = mpz_fdiv_ui(M->entry[i][j], (unsigned long)p);
    detp = FFdet(Mp);
    detp = FFdiv(detp, Dp);
    det_changed = mpz_cra_ui(det, det_modulus, detp, p);
    mpz_mul_ui(det_modulus, det_modulus, (unsigned long)p);
    log_det_modulus += logi((unsigned long)p);
    FFdtor(FFp);
  }
  mpz_clear(det_modulus);
  PrimeSourceDtor(PS);
  FFmat_dtor(Mp);
  }
printf("Factor from CRA:            %8.1f bits\n", mpz_log(det)/log(2.0));
  mpz_mul(det, det, det_factor);
printf("Final answer:               %8.1f bits\n", mpz_log(det)/log(2.0));
  /* juggle sign in case a row/column swap was not really a swap */
  if ((row == n-1) ^ (col == n-1)) mpz_neg(det, det);
tidy_up:
  mpz_clear(det_factor);
  mpz_clear(D);
  /* Put matrix M back as it was originally */
  Zmat_swap_rows(M, row, n-1);
  Zmat_swap_cols(M, col, n-1);
}


