//   Copyright (c)  1997-2006  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License (version 2)
//   as published by the Free Software Foundation.  A copy of the full
//   licence may be found in the file COPYING in this directory.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoA; if not, write to the Free Software
//   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


#include <stdio.h>
#include <math.h>
// the line below for debugging only...
#include "DUPFFprint.h"
#include "DUPFFbivariate_lift2.h"
#include "jalloc.h"
#include "jaaerror.h"
#include "jtime.h"



/***************************************************************************/
/* Conversion of poly from a y-adic expansion to a (y-alpha)-adic expansion*/
/* Input is poly  = F[0] +        y *F[1] + ... +         y^deg * F[deg].  */
/* On return poly = F[0] + (y-alpha)*F[1] + ... + (y-alpha)^deg * F[deg].  */


void DUPFFbiv_shift(const DUPFF *F, int deg, int alpha)
{
  int i, j;

  for (i=0; i < deg; i++)
    for (j=deg-1; j >= i; j--)
      DUPFFshift_add(F[j],F[j+1],0,alpha);/* F[j] += alpha * F[j+1] */
}

/***************************************************************************/
/* Lift a univariate factorization to a bivariate one.                     */
/* Input is f a y-adic expansion of the bivariate polynomial,              */
/* and the univariate factorization of f at y=alpha.                       */
/* Output is array such that lifted_factors[i] is y-adic expn of lifted factor*/

DUPFF** DUPFFbivariate_lift(DUPFF *f, int dy, DUPFFlist factors, uint32 alpha)
{
  DUPFF **lifted_factors;
  DUPFFbiv_lifter lifter;
  int i, nfactors, deg;
  int p = CurrentFF.prime;

printf("shifting...\n");print_time();
  if (alpha != 0)
    DUPFFbiv_shift(f, dy, alpha); /* convert to (y-alpha)-adic expansion */
printf("shifted.\n");print_time();
  lifter = DUPFFbiv_lift_init(f, dy, factors);
  for (i=1; i <= dy/2; i++)
  {if(i%100==0){printf("level=%d\n",i);print_time();}
    DUPFFbiv_lift_step(lifter, f[i]);
  }
printf("lifting completed (there may be 1 bogus factor)\n");print_time();
  lifted_factors = (DUPFF **)MALLOC(2*sizeof(DUPFF *));
  nfactors = DUPFFbiv_lift_output(lifted_factors, lifter, 0);
  DUPFFbiv_lifter_dtor(lifter);
printf("factors output\n");print_time();  
  if (alpha != 0)
    for (i=0; i < nfactors; i++)
    {
      for (deg=dy/2; deg > 0; deg--)
        if (DUPFFdeg(lifted_factors[i][deg]) >= 0) break;
      DUPFFbiv_shift(lifted_factors[i], deg, p-alpha);
    }
printf("factors shifted\n");print_time();  
  return lifted_factors;
}



/***************************************************************************/

void solve_for_correction(DUPFF *gdelta, const DUPFF E, const DUPFF g0, const DUPFF grecip)
{
  DUPFF junk, tmp;
  int dg0, dE;

  dg0 = DUPFFdeg(g0);
  dE = DUPFFdeg(E);
  junk = DUPFFnew(dE + dg0);  /* these are all a bit too big */
  tmp = DUPFFnew(2*dg0);      /*                             */
  *gdelta = DUPFFnew(dg0);    /*                             */

  DUPFFdiv4(junk, tmp, E, g0);
  DUPFFmul3(tmp, tmp, grecip);
  DUPFFdiv4(junk, *gdelta, tmp, g0);

  DUPFFfree(tmp);
  DUPFFfree(junk);
}

/***************************************************************************/


DUPFFbiv_lifter DUPFFbiv_lifter_ctor(DUPFF g, DUPFF h, DUPFFbiv_lifter g_lifter, DUPFFbiv_lifter h_lifter, int rmax)
{
  DUPFFbiv_lifter ans;
  DUPFF grecip, hrecip, tmp;
  int df;
  
//  if (DUPFFdeg(g) > DUPFFdeg(h)) return DUPZfactor_lifter_ctor(h, g, h_lifter, g_lifter);
  ans = (DUPFFbiv_lifter)MALLOC(sizeof(struct DUPFFbiv_lift_struct));
  ans->r = 1;
  ans->g = (DUPFF*)MALLOC(rmax*sizeof(DUPFF));
  ans->h = (DUPFF*)MALLOC(rmax*sizeof(DUPFF));
  ans->g[0] = DUPFFcopy(g);
printf("init g[0] to ");DUPFFprint(ans->g[0]);
  ans->h[0] = DUPFFcopy(h);
  df = DUPFFdeg(g) + DUPFFdeg(h);
  ans->E = DUPFFnew(df);
  
  tmp = DUPFFexgcd(&grecip, &hrecip, g, h);
  if (DUPFFdeg(tmp) != 0) { JERROR(JERROR_HENSEL); exit(1); }
  /* since the gcd is not monic we have to do this... */
  DUPFFdiv2ff(grecip, DUPFFlc(tmp));
  DUPFFdiv2ff(hrecip, DUPFFlc(tmp));
  ans->grecip = grecip;
  ans->hrecip = hrecip;
  ans->g_lifter = g_lifter;
  ans->h_lifter = h_lifter;
  DUPFFfree(tmp);
  return ans;
}



void DUPFFbiv_lifter_dtor(DUPFFbiv_lifter this)
{
//  int i;
  int r;

  if (this == NULL) return;
  r = this->r;
  DUPFFfree(this->E);
  if (this->g_lifter) DUPFFbiv_lifter_dtor(this->g_lifter);
  if (this->h_lifter) DUPFFbiv_lifter_dtor(this->h_lifter);
//  for (i=0; i < r; i++)
//  {
//    DUPFFfree(this->g[i]);
//    DUPFFfree(this->h[i]);
//  }
//  FREE(this->g);
//  FREE(this->h);
  DUPFFfree(this->grecip);
  DUPFFfree(this->hrecip);
  FREE(this);
}

/***************************************************************************/

void DUPFFbiv_lift_step(DUPFFbiv_lifter this, DUPFF f_cmpt_r)
{
  int i;
  /* the rest are just aliases to make the code easier to read (!?!) */
  DUPFF       *g     = this->g;
  DUPFF       *h     = this->h;
  int         r      = this->r;
  const DUPFF grecip = this->grecip;
  const DUPFF hrecip = this->hrecip;
  DUPFF       E      = this->E;
DUPFF tmp;

  tmp = DUPFFnew(DUPFFdeg(g[0]) + DUPFFdeg(h[0]));
  DUPFFcopy2(E, f_cmpt_r);
  for (i=1; i < r; i++)
  {
    DUPFFmul3(tmp, g[i], h[r-i]);
    DUPFFsub3(E, E, tmp);
  }
  DUPFFfree(tmp);

//printf("before solve g[0]=");DUPFFprint(g[0]);
  solve_for_correction(&g[r], E, g[0], hrecip);
//printf("after solve g[0]=");DUPFFprint(g[0]);
//printf("before solve h[0]=");DUPFFprint(h[0]);
  solve_for_correction(&h[r], E, h[0], grecip);
//printf("after solve h[0]=");DUPFFprint(h[0]);
  this->r++;

  if (this->g_lifter) DUPFFbiv_lift_step(this->g_lifter, g[r]);
  if (this->h_lifter) DUPFFbiv_lift_step(this->h_lifter, h[r]);
}


/***************************************************************************/
/* Convert the p-adic expansions of the lifted factors to their normal     */
/* representations.                                                        */

int DUPFFbiv_lift_output(DUPFF **factors, const DUPFFbiv_lifter this, int i)
{
  int p, k;

  p = CurrentFF.prime;
  k = this->r;
  if (this->g_lifter) i = DUPFFbiv_lift_output(factors, this->g_lifter, i);
  else factors[i++] = this->g;
  if (this->h_lifter) i = DUPFFbiv_lift_output(factors, this->h_lifter, i);
  else factors[i++] = this->h;
  return i;
}



/***************************************************************************/

static DUPFFbiv_lifter DUPFFbiv_lift_init_stupid(const DUPFFlist pfactors, int rmax)
{
  int df;
  DUPFFbiv_lifter ans;
  DUPFF tmp;
  DUPFFlist iter;

  df = 0;
printf("factors are:\n");
  for (iter = pfactors; iter; iter = iter->next)
  {
    df += DUPFFdeg(iter->poly);
DUPFFprint(iter->poly);
  }
  ans = NULL;
  tmp = DUPFFnew(df);
  DUPFFcopy2(tmp, pfactors->poly);
  for (iter = pfactors->next; iter; iter = iter->next)
  {
    ans = DUPFFbiv_lifter_ctor(iter->poly, tmp, NULL, ans, rmax);
    DUPFFmul3(tmp, tmp, iter->poly);
  }
  DUPFFfree(tmp);
  return ans;
}


DUPFFbiv_lifter DUPFFbiv_lift_init(DUPFF *f, int dy, DUPFFlist pfactors)
{
  return DUPFFbiv_lift_init_stupid(pfactors, dy);
}



#if 0
///****************************************************************************/
///* This function is called after a speculative early search which has found */
///* at least one factor but further lifting is still required.               */
///* It creates a new lifting tree for the remaining factors.                 */
//
//void DUPFFbiv_lift_revise1(DUPFFfactor_info info, int height)
//{
//  int i;
//
//  DUPFFbiv_lift_init(info);
//  DUPFFpadic_expand_end(info->f_expansion);
//  info->f_expansion = DUPFFpadic_expand_init(info->f, info->p, info->rmax);
//  DUPFFpadic_expand_step(info->f_expansion);
//  if (height > info->rmax) height = info->rmax;
//  for(i=1; i < height; i++)
//  {
//    DUPFFpadic_expand_step(info->f_expansion);
//    DUPFFbiv_lift_step(info->lifter, info->f_expansion->monic_cmpts[i]);
//  }
//}
//
//
//void DUPFFbiv_lift_revise(DUPFFfactor_info info)
//{
//  int height = info->lifter->r;
//
//  DUPFFbiv_lift_dtor(info->lifter); info->lifter = NULL;
//  DUPFFpadic_expand_end(info->f_expansion); info->f_expansion = NULL;
//  DUPFFbiv_lift_revise1(info, height);
//}
#endif

/***************************************************************************/
void liftprint(DUPFFbiv_lifter this, int indent)
{
int i;
if (this == NULL) return;
putchar('>');for(i=0;i<indent;i++)putchar(' ');
for(i=0;i<this->r;i++)printf("%d,",DUPFFdeg(this->g[i]));printf("\n");
for(i=0;i<=indent;i++)putchar(' ');
for(i=0;i<this->r;i++)printf("%d,",DUPFFdeg(this->h[i]));printf("\n");
liftprint(this->g_lifter, indent+2);
liftprint(this->h_lifter, indent+2);
}

